--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

local UnitHasVehicleUI = UnitHasVehicleUI or function() end
local UnitGroupRolesAssigned = UnitGroupRolesAssigned or function() end

local lower, strlen, strsub, strfind = string.lower, string.len, string.sub, string.find
local wipe = table.wipe

local UnitWidth, UnitHeight, xOffset, yOffset
local L, PlayerName, PlayerColor, Settings
local LeaderHighlight, PlayerHighlight

local DebuffCount = 3
local Font = NS.Font
local MaxAuras = 40
local TexturePath = NS.TexturePath .. "RaidFrames/"
local ThresholdLow = 0.6
local ThresholdCritical = 0.3
local UpdateSpeedBase = 0.1  -- 10 updates per second (roughly)

local DebuffList = {}
local DefferUpdate = {}
local GroupHeaders = {}
local PowerList = {}
local UnitsEntries = {}
local UnitEventHandlers = {}
local BackdropCold = { bgFile="Interface\\BUTTONS\\WHITE8X8", edgeFile=TexturePath.."UI-Tooltip-Border", tile=true, edgeSize=8, insets={ left=2, right=2, top=2, bottom=2 } }
--local BackdropHot = { bgFile="Interface\\BUTTONS\\WHITE8X8", edgeFile=TexturePath.."UI-Tooltip-BorderCold", tile=true, edgeSize=8, insets={ left=2, right=2, top=2, bottom=2 } }
local RoleList = {
	DAMAGER={ tip="Damage", icon="damage", size=12, },
	HEALER={ tip="Healer", icon="healer", size=12, },
	TANK={ tip="Tank", icon="tank", size=14, },
}

local WatchFrame = CreateFrame("Frame")

local function GetHealthColor(self, maxhealth, health, thresholdlow, thresholdcritical) -- Recolours the health bar as it gets low
	if not self.DefaultColor then
		return
	end
	local percent
	if health and maxhealth then
		if maxhealth == 0 then
			percent = health
		else
			percent = health/maxhealth
		end
	else
		return self.DefaultColor.r, self.DefaultColor.g, self.DefaultColor.b
	end
	local r, g, b, diff
	if percent <= thresholdcritical then
		r = self.CriticalColor.r
		g = self.CriticalColor.g
		b = self.CriticalColor.b
	elseif percent <= thresholdcritical then
		diff = 1 - (percent - thresholdcritical) / (thresholdlow - thresholdcritical)
		r = self.LowColor.r - (self.LowColor.r - self.CriticalColor.r) * diff
		g = self.LowColor.g - (self.LowColor.g - self.CriticalColor.g) * diff
		b = self.LowColor.b - (self.LowColor.b - self.CriticalColor.b) * diff
	elseif percent < 1 then
		diff = 1 - (percent - thresholdlow) / (1 - thresholdlow)
		r = self.DefaultColor.r - (self.DefaultColor.r - self.LowColor.r) * diff
		g = self.DefaultColor.g - (self.DefaultColor.g - self.LowColor.g) * diff
		b = self.DefaultColor.b - (self.DefaultColor.b - self.LowColor.b) * diff
	else
		return self.DefaultColor.r, self.DefaultColor.g, self.DefaultColor.b
	end
	if r < 0 then r = 0; end
	if r > 1 then r = 1; end
	if g < 0 then g = 0; end
	if g > 1 then g = 1; end
	if b < 0 then b = 0; end
	if b > 1 then b = 1; end
	return r, g, b
end

local function SetOrder()
--	if Settings.orderbygroup == true then
		GroupHeaders[1]:SetAttribute("groupFilter", "1,3,5,7")
		GroupHeaders[2]:SetAttribute("groupFilter", "2,4,6,8")
		GroupHeaders[1]:SetAttribute("nameList",  nil)
 		GroupHeaders[1]:SetAttribute("sortMethod",  nil)
		GroupHeaders[2]:SetAttribute("nameList",  nil)
 		GroupHeaders[2]:SetAttribute("sortMethod",  nil)
--	elseif
--	else
--	end
end

local function SetUnitBackdrop(self)
	self:SetBackdrop(BackdropCold)
end

local function SetUnitColors(self)
	self:SetBackdropColor(Settings.backgroundcolor.r, Settings.backgroundcolor.g, Settings.backgroundcolor.b, Settings.backgroundcolor.a)	
	self.Health.DefaultColor = Settings.healthcolor.default
	self.Health.LowColor = Settings.healthcolor.low
	self.Health.CriticalColor = Settings.healthcolor.critical
end

local function SetNameColor(self, alpha)
	local frame, unitid
	if self.Frame then
		frame = self.Frame
		unitid = self.UnitID
	else
		frame = self
		unitid = self:GetParent().UnitID
	end
	if UnitHasVehicleUI(unitid) then		-- Vehicle Support
		frame.NameText:SetTextColor(0, .5, 0, alpha)
	else
		if not Settings.classnamecolors then
			frame.NameText:SetTextColor(Settings.namecolor.r, Settings.namecolor.g, Settings.namecolor.b, alpha)
			frame.NameText:SetShadowColor(Settings.nameshadow.r, Settings.nameshadow.g, Settings.nameshadow.b, alpha) 
		else
			local class = select(2, UnitClass(unitid)) or "ROGUE" -- My main is a rogue so go figure
			local Color = RAID_CLASS_COLORS[class]
			frame.NameText:SetTextColor(Color.r, Color.g, Color.b, alpha)
			frame.NameText:SetShadowColor(0, 0 , 0, alpha) 
		end
	end
end

local function CreateHighlight(texture, drawlayer, level)
	local f = CreateFrame("Frame", nil, UIParent)
	f:SetSize(10, 10)
	f:SetFrameStrata("MEDIUM")
	f.Texture = f:CreateTexture()
	level = level or 1
	if drawlayer then
		f.Texture:SetDrawLayer(drawlayer, level)
	end
	f.Texture:SetAllPoints()
	f.Texture:SetBlendMode("ADD")
	f.Texture:SetTexture(texture)
	f:Hide()
	return f
end

local function SetSelfHighlight(frame, unitid)
	local IsPlayer
	local name, realm = UnitName(unitid)
	if name == PlayerName and not realm then
		IsPlayer = true
	end
	local Leader = UnitIsGroupLeader(unitid)
	if Leader and IsPlayer then
		PlayerHighlight:Hide()
	end
	if Settings.highlightleader and Leader then
		LeaderHighlight:SetParent(frame)
		LeaderHighlight:ClearAllPoints()
		LeaderHighlight:SetPoint("BOTTOMLEFT", frame)
		return
	end
	if Settings.highlightplayer and IsPlayer then
		PlayerHighlight:SetParent(frame)
		PlayerHighlight.Texture:SetVertexColor(PlayerColor.r, PlayerColor.g, PlayerColor.b)
		PlayerHighlight:ClearAllPoints()
		PlayerHighlight:SetPoint("BOTTOMLEFT", frame)
		if not PlayerHighlight:IsShown() then
			PlayerHighlight:Show()
		end
	end
end

local function UpdateRaidIcon(self, unitid)
	local targetIndex = GetRaidTargetIndex(unitid.."target")
	if not Settings.raidicons or not targetIndex then
		if self.RaidIcon:IsShown() then
			self.RaidIcon:Hide()
		end
		return
	end
	if targetIndex == 1 or targetIndex == 4 or targetIndex == 5 or targetIndex == 6 then -- 1=star, 2=circle, 3=diamond, 4=triangle or 5=moon, 6=square, 7=cross, 8=skull
		self.RaidIcon:SetAlpha(Settings.raidiconalpha + .2)
	else
		self.RaidIcon:SetAlpha(Settings.raidiconalpha)
	end
	SetRaidTargetIconTexture(self.RaidIcon, targetIndex)
	self.RaidIcon:Show()
end

local function UpdateSettings()
	wipe(PowerList)
	ThresholdLow = Settings.healthcolor.thresholdlow
	ThresholdCritical = Settings.healthcolor.thresholdcritical
	PlayerHighlight:Hide()
	LeaderHighlight:Hide()
	if Settings.highlightplayer then
		PlayerHighlight:Show()
	end
	if Settings.highlightleader then
		LeaderHighlight:Show()
	end
	for key,_ in pairs(RoleList) do 
		RoleList[key].showrole = Settings.showrole[key]
	end
	for k, v in pairs(Settings.showpower) do 
		if v then
			PowerList[k] = true
		end
	end
	for f,_ in pairs(UnitsEntries) do
		local unitid = f:GetParent().UnitID
		if unitid then
			SetUnitBackdrop(f)
			SetUnitColors(f)
			SetNameColor(f, 1)
			UpdateRaidIcon(f, unitid)
			UnitEventHandlers.UNIT_HEALTH(f:GetParent(), f:GetParent().UnitID)
			UnitEventHandlers.UNIT_DISPLAYPOWER(f:GetParent(), f:GetParent().UnitID)
		end
	end
end
NS:AddOption("UpdateRaidSettings", UpdateSettings)

local function ShowTooltip(frame)
	local unitid = frame.UnitID
	GameTooltip:ClearLines()
	GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
	GameTooltip:SetUnit(unitid)
	local _, _, num = unitid:find("(%d+)")
	if num then
		local _, _, subgroup = GetRaidRosterInfo(num)
		GameTooltip:AddLine(L.Group .. ": " .. subgroup) 
	end
	role = UnitGroupRolesAssigned(unitid)
	if role and role ~= "NONE" then 
		GameTooltip:AddLine(L.Role .. ": " .. RoleList[role].tip) 
		GameTooltip:AddTexture(TexturePath .. RoleList[role].icon)
	end
	GameTooltip:Show()
end

local function HideTooltip()
	GameTooltip:FadeOut()
end

local function RangeUpdate(self, elapsed)
	local unitid = self.UnitID
	if not unitid then return end
	if not self.timer then 
		self.timer = UpdateSpeedBase 
		self.updatespeed = self.timer
	end
	self.timer = self.timer - elapsed
	if self.timer > 0 then return end
	self.timer = self.updatespeed

	-- Range Check
	local inRange,_ = UnitInRange(unitid)
	if (not IsInGroup()) or inRange then 
		self.Frame:SetAlpha(1)
	else
		self.Frame:SetAlpha(Settings.oorangealpha) 
	end
end

local function OnEvent(self, event, ...)
	UnitEventHandlers[event](self, ...) 
end

local function SetHandlers(self)
	self:SetScript("OnShow", OnShowUnit)
	self:SetScript("OnHide", OnHideUnit)
	self:SetScript("OnEvent", OnEvent)
	self:SetScript("OnUpdate", RangeUpdate)
	for eventname, _ in pairs(UnitEventHandlers) do 
		self:RegisterEvent(eventname) 
	end
end

local function InitClicks(self)
	if self.attrSet then return end
	if InCombatLockdown() then
		DefferUpdate[self] = true
		return
	end
	self.attrSet = true
	self:RegisterForClicks("AnyUp")
	self:SetAttribute("*type1","target")
	self:SetAttribute("*type2", "togglemenu")
	if FocusFrame then -- TBC+
		self:SetAttribute("toggleForVehicle", true) -- Vehicle Support
	end
	SetHandlers(self)
end

local function DebuffOnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_LEFT")
	local unit = self.debuffInfo.unitid
	local debuffindex = self.debuffInfo.debuffindex
	GameTooltip:ClearLines()
	if unit=="player" then
		GameTooltip:SetUnitDebuff("player", debuffindex, "HARMFUL")
	else
		GameTooltip:SetUnitDebuff(unit, debuffindex)
	end
	GameTooltip:Show()
end

local function DebuffOnLeave()
	GameTooltip:Hide()
end

local function ApplyDebuff(self)
	local debuffInfo = self.debuffInfo
	local color, text, textcolor
	if debuffInfo.icon then
		self.Image:SetTexture(debuffInfo.icon)
		self:Show()
	else
		self.Image:SetTexture("")
		self.debuffInfo = {}
		self:Hide()
	end
end

local function SetDebuffWatch()
	WatchFrame:SetScript("OnUpdate", WatchDebuffs)
end

local function UpdateDebuffs(debuffs, unitid)
	local debuffCounter = 0
	local debuff
	for i = 1, MaxAuras do
--[[
LightSpark:
To get only dispellable debuffs, you need to pass different 3rd argument to UnitAura(). It's even mentioned on wowpedia.

RAID - auras that can be applied (if HELPFUL) or dispelled (if HARMFUL) by the player.
So, your third arg should be "HARMFUL|RAID" instead of just "HARMFUL". And if you know that the default UI does it, you should look for it in Blizz UI code then, it's fairly easy to find.
]]--
-- Fizz		local name, rank, icon, count, dispelType , duration, expirationTime, isMine, isStealable, shouldConsolidate, spellID, canApplyAura, isBossDebuff = UnitDebuff(unitid, i)
		local name, icon, count, dispelType , duration, expirationTime, isMine, isStealable, shouldConsolidate, spellID, canApplyAura, isBossDebuff = UnitDebuff(unitid, i)
		if not name then 
			if debuffCounter < DebuffCount then
				for debuffCounter = debuffCounter + 1,  DebuffCount do
					debuff = debuffs[debuffCounter]
					debuff.debuffInfo = {}
					DebuffList[debuff] = nil
					ApplyDebuff(debuff)
				end
			end
			return 
		end
		if icon and dispelType then
			debuffCounter = debuffCounter + 1
			if debuffCounter > DebuffCount then return end
			if not expirationTime then
				expirationTime = 0
			end
			debuff = debuffs[debuffCounter]
			debuff.debuffInfo = {}
			debuff.debuffInfo.icon = icon
			debuff.debuffInfo.debuffs = debuffs
			debuff.debuffInfo.unitid = unitid
			debuff.debuffInfo.spellid = spellID
			debuff.debuffInfo.debuffindex = i 
			ApplyDebuff(debuff)
			if #DebuffList == 0 then
				SetDebuffWatch()
			end
			DebuffList[debuff] = expirationTime
		end
	end
end

local function WatchDebuffs(self, elapsed)
	if not self.timer then 
		self.timer = UpdateSpeedBase 
		self.updatespeed = self.timer
	end
	self.timer = self.timer - elapsed
	if self.timer > 0 then return end
	self.timer = self.updatespeed
	local time = GetTime()
	local count, nullcount = 0, 0
	local NullFrames = {}
	for frame, expiration in pairs(DebuffList) do
		if expiration < time then 
			NullFrames[frame.debuffInfo.unitid] = frame.debuffInfo.debuffs
			nullcount = nullcount + 1
			DebuffList[frame] = nil
			frame.debuffInfo = {}
		else
			count = count + 1 
		end
		ApplyDebuff(frame) -- Hide will happen here
	end
	-- If no more frames to watch, unregister the OnUpdate script
	if count == 0 then
		self:SetScript("OnUpdate", nil)
		self.timer = nil
	end
	 -- unit has a spare debuff frame so check to see if there are more than debuffCount on the unit
	if nullcount then
		for unitid, debuffs in pairs(NullFrames) do
			UpdateDebuffs(debuffs, unitid)
		end
	end
end

local function UpdateUnit(self, unitid)
	self.Nextupdate = 0
	local frame = self.Frame
	-- Name
	frame.NameText:SetText(UnitName(unitid))
	SetSelfHighlight(frame, unitid, Settings)
	-- Health
	local maxhealth = UnitHealthMax(unitid)
	frame.Health:SetMinMaxValues(0,maxhealth)
	
	-- Power		
	UnitEventHandlers.UNIT_DISPLAYPOWER(self, unitid)
	UnitEventHandlers.UNIT_AURA(self, unitid)
	UnitEventHandlers.PARTY_MEMBER_DISABLE(self, unitid)
	UnitEventHandlers.UNIT_HEALTH(self, unitid)
--	UnitEventHandlers.INCOMING_RESURRECT_CHANGED(self, unitid)
	if UnitGetIncomingHeals then
	-- Incoming Heals
		frame.IncomingHeal:SetMinMaxValues(0,maxhealth)
		UnitEventHandlers.UNIT_HEAL_PREDICTION(self, unitid)	
	end
	--UnitEventHandlers.RAID_TARGET_UPDATE(self)
	-- Role
	local role = UnitGroupRolesAssigned(unitid)
	if not role or role == "NONE" or not RoleList[role].showrole then
		frame.RoleIcon:Hide()
	else
		frame.RoleIcon:SetWidth(RoleList[role].size)
		frame.RoleIcon:SetHeight(RoleList[role].size)
		frame.RoleIcon:SetTexture(TexturePath .. RoleList[role].icon)
		frame.RoleIcon:Show()
	end
	UpdateRaidIcon(frame, unitid)
--	SetUnitColors(f, Settings)
	SetNameColor(frame, 1, Settings)
end

local function OnShowUnit(self)
	InitClicks(self)
	local unitid = self:GetAttribute("unit")
	if unitid then 
		self.UnitID = unitid
		UpdateUnit(self, unitid)
	end
end

local function OnHideUnit(self)
	self.attrSet = false
	self:SetScript("OnUpdate", nil)
end

local function OnChangeAttribute(self, attribute, value)
	if attribute == "unit" then 
		InitClicks(self)
		local unitid = self:GetAttribute("unit")
		if unitid then 
			self.UnitID = unitid
			UpdateUnit(self, unitid) 
		end
	end
end

function UnitEventHandlers.UNIT_AURA(self, unit)
	local unitid = self.UnitID
	if not unitid or unit ~= unitid then return end -- Changed
	UpdateDebuffs(self.Frame.Debuffs, unitid)
end

--[[
function UnitEventHandlers.INCOMING_RESURRECT_CHANGED(self, unit)
	local unitid = self.UnitID
	if not unitid or unit ~= unitid then return end
	if not UnitHasIncomingResurrection(unit) then
		self.ResurrectIcon:SetTexture(nil)
	else
		self.ResurrectIcon:SetTexture("Interface\\RaidFrame\\Raid-Icon-Rez")
	end
end
]]--
function UnitEventHandlers.UNIT_DISPLAYPOWER(self, unit)
	local unitid = self.UnitID
	if not unitid then return end
	if unit == unitid then 
		local _, powertype = UnitPowerType(unitid or "")
		if PowerList[powertype] then
			self.Frame.Power:Show()
			self.Frame.Power.show = true
			self.Frame.Power:SetMinMaxValues(0,UnitPowerMax(unitid))
			self.Frame.Power:SetValue(UnitPower(unitid))
		else
			self.Frame.Power.show = nil
			self.Frame.Power:Hide()
		end
	end
end

function UnitEventHandlers.UNIT_POWER_UPDATE(self, unit, token) -- UNIT_POWER_UPDATE only fires every 2 second unlike the lots of UNIT_POWER_FREQUENT
	local unitid = self.UnitID
	if not unitid then return end
	if unit == unitid then 
		self.Frame.Power:SetValue(UnitPower(unitid))
	end
end
UnitEventHandlers.UNIT_MAXPOWER = UnitEventHandlers.UNIT_POWER_UPDATE

function UnitEventHandlers.UNIT_TARGET(self, unit)
	local unitid = self.UnitID
	if not unitid then return end
	if unit == unitid then 
		UpdateUnit(self, self.UnitID) 
	end
end
UnitEventHandlers.RAID_TARGET_UPDATE = UnitEventHandlers.UNIT_TARGET

function UnitEventHandlers.PARTY_MEMBER_DISABLE(self, unit)
	local unitid = self.UnitID
	if not unitid or unit ~= unitid then return end
	local text = ""
	if UnitIsDead(unitid) then text = L.Dead end
	if UnitIsGhost(unitid) then text = L.Released end
	if not UnitIsConnected(unitid) then text = L.Disconnected end
	if text ~= "" then 
		self.Frame:SetBackdropColor(0, 0, 0, Settings.backgroundcolor.a)
		self.Frame.Health:SetAlpha(0)
		if self.Frame.Power.show then
			self.Frame.Power:SetAlpha(0)
		end
		self.Frame.DeadText:SetText(text)
		self.Frame.NameText:SetTextColor(1, 1, 1, .45) --dead
		self.Frame.DeadText:Show()
	else
		self.Frame:SetBackdropColor(Settings.backgroundcolor.r, Settings.backgroundcolor.g, Settings.backgroundcolor.b, Settings.backgroundcolor.a)	
		self.Frame.Health:SetAlpha(1)
		if self.Frame.Power.show then
			self.Frame.Power:SetAlpha(1)
		end
		SetNameColor(self, 1, Settings)
		self.Frame.DeadText:Hide()
	end
end
UnitEventHandlers.PARTY_MEMBER_ENABLE = UnitEventHandlers.PARTY_MEMBER_DISABLE
UnitEventHandlers.PLAYER_DEAD = UnitEventHandlers.PARTY_MEMBER_DISABLE

function UnitEventHandlers.UNIT_HEALTH(self, eventunit)
	local unitid = self.UnitID
	if not unitid then return end
	local petunit, visibleunit
	if unitid == "player" then 
		petunit = "pet" 
	else
		petunit = unitid.."pet" 
	end
	if eventunit == unitid or eventunit == petunit then 
		if UnitHasVehicleUI(unitid) then 
			visibleunit = petunit
		else
			visibleunit = unitid 
		end
		if UnitExists(visibleunit) then
			-- Sets Health Bar
			local MaxHealth = UnitHealthMax(visibleunit)
			local Health = UnitHealth(visibleunit)
			self.Frame.Health:SetMinMaxValues(0, MaxHealth)
			self.Frame.Health:SetValue(Health)	
			local r, g, b = GetHealthColor(self.Frame.Health, MaxHealth, Health, ThresholdLow, ThresholdCritical)
			self.Frame.Health:SetStatusBarColor(r, g, b, 1)
			self.Frame.IncomingHeal:SetStatusBarColor(r, g, b, .5)
		end
	end
end
UnitEventHandlers.UNIT_MAXHEALTH = UnitEventHandlers.UNIT_HEALTH
if NS:IsValidVersion(NS.WOW_Project.Classic) then
	UnitEventHandlers.UNIT_HEALTH_FREQUENT = UnitEventHandlers.UNIT_HEALTH
end

function UnitEventHandlers.PLAYER_TARGET_CHANGED(self) 
	local unitid = self.UnitID
	if not unitid then return end
	local visibleunit
	if UnitHasVehicleUI(unitid) then
		if unitid == "player" then
			visibleunit = "pet" 
		else
			visibleunit = unitid.."pet" 
		end
	else
		visibleunit = unitid 
	end
end

function UnitEventHandlers.UNIT_THREAT_SITUATION_UPDATE(self, unit)
	local unitid = self.UnitID
	if not unitid then return end
	local unitaggro = UnitThreatSituation(unitid)
	if unitaggro and unitaggro > 1 then 
		self.Frame.Aggro:Show() 
	else
		self.Frame.Aggro:Hide() 
	end
end

function UnitEventHandlers.UNIT_PORTRAIT_UPDATE(self, unit)
	local unitid = self.UnitID
	if not unitid then return end
	UnitEventHandlers.UNIT_DISPLAYPOWER(self, unitid)
	UnitEventHandlers.UNIT_AURA(self, unitid)
	UnitEventHandlers.PARTY_MEMBER_DISABLE(self, unitid)	
	UnitEventHandlers.UNIT_HEALTH(self, unitid)
end
UnitEventHandlers.PLAYER_UNGHOST = UnitEventHandlers.UNIT_PORTRAIT_UPDATE
UnitEventHandlers.PLAYER_ALIVE = UnitEventHandlers.UNIT_PORTRAIT_UPDATE

function UnitEventHandlers.UNIT_FLAGS(self, unit)
	local unitid = self.UnitID
	if not unitid then return end
	UnitEventHandlers.UNIT_AURA(self, unitid)
	UnitEventHandlers.PARTY_MEMBER_DISABLE(self, unitid)
	
end

function UnitEventHandlers.UNIT_PET(self, unit)
	OnChangeAttribute(self, "unit")
	if self.UnitID then 
		UpdateUnit(self, self.UnitID) 
	end
end

if UnitGetIncomingHeals then
	UnitEventHandlers.UNIT_ENTERED_VEHICLE = UnitEventHandlers.UNIT_PET
	UnitEventHandlers.UNIT_EXITED_VEHICLE = UnitEventHandlers.UNIT_PET

	function UnitEventHandlers.UNIT_HEAL_PREDICTION(self, unit)
		local unitid = self.UnitID
		if not unitid then return end

		if unit == unitid then
			local incomingHeal = UnitGetIncomingHeals(unitid) or 0
			if incomingHeal > 0 then incomingHeal = UnitHealth(unitid)+incomingHeal end
			self.Frame.IncomingHeal:SetValue(incomingHeal)	
		end
	end
end

local function CreateUnit(self)
	-- Create Base Frames
	self.AuraWatcher = CreateFrame("Frame", "$parent_Watcher", self)
	local f = CreateFrame("Frame", self:GetName().."Overlay", self, BackdropTemplateMixin and "BackdropTemplate")
	self.Frame = f
	self.auraUpdateTime = 0

	UnitsEntries[f] = { ["index"]=self:GetParent().index }
	f:SetPoint("CENTER", self)
	f:SetWidth(UnitWidth)
	f:SetHeight(UnitHeight)
--	f:SetBackdropBorderColor(.8, .8, .8)

	-- Health Bar 
	f.Health = CreateFrame("StatusBar", "$parent_Health", f, BackdropTemplateMixin and "BackdropTemplate")
	f.Health:SetStatusBarTexture(TexturePath.."runebar")

	f.Health:SetBackdrop({bgFile = TexturePath.."bar"})
	f.Health:SetBackdropColor(1, 1, 1, .45)

	f.Health:SetOrientation("HORIZONTAL")
	f.Health:SetPoint("LEFT", 3, -4)
	f.Health:SetFrameLevel(f:GetFrameLevel()+1)
	f.Health:SetWidth(UnitWidth-6)
	f.Health:SetHeight((UnitHeight/2)-3)

	-- Heal Prediction Bar 
	f.IncomingHeal = CreateFrame("StatusBar", nil, f.Health)
	f.IncomingHeal:SetStatusBarTexture(TexturePath.."runebar")
	f.IncomingHeal:SetOrientation("HORIZONTAL")
	f.IncomingHeal:SetAllPoints(f.Health)
	f.IncomingHeal:SetStatusBarColor(0, 1, 0, .6)
	
	-- Overlay Anchor (To make layering easier)
	f.OverlayAnchor = CreateFrame("StatusBar", nil, f.Health)
	f.OverlayAnchor:SetAllPoints()
	f.OverlayAnchor:SetFrameLevel(f.IncomingHeal:GetFrameLevel()+1)

	-- Power (Mana) Bar
	f.Power = CreateFrame("StatusBar", nil, f.OverlayAnchor, BackdropTemplateMixin and "BackdropTemplate")
	f.Power:SetStatusBarTexture(TexturePath.."Bar")
--	f.Power:SetBackdrop({bgFile = TexturePath.."Bar"});
	f.Power:SetBackdrop({ bgFile = "Interface\\BUTTONS\\WHITE8X8", tile=true, });
	f.Power:SetOrientation("HORIZONTAL")
	f.Power:SetPoint("TOPLEFT", f.Health, "BOTTOMLEFT")
	f.Power:SetWidth(f.Health:GetWidth())		
	f.Power:SetHeight(5)	
	f.Power:SetBackdropColor(0, 0, 0, 1)
	f.Power:SetStatusBarColor(.419, .968, 1, 1)  -- FM Mana color

	-- Create Name Text Field
	f.NameText = f:CreateFontString(nil, "BACKGROUND")
	f.NameText:SetFont(Font, 11, "")
	f.NameText:SetPoint("TOP", f, "TOP", 0, -2)
	f.NameText:SetWidth(f:GetWidth())
	f.NameText:SetHeight(10)
	f.NameText:SetShadowColor(0, 0 , 0, 1) 
	f.NameText:SetShadowOffset(1.2, -1.2)
	f.NameText:SetJustifyV("TOP")
	f.NameText:SetJustifyH("CENTER")

	-- Create Dead/Released Text
	f.DeadText = f:CreateFontString(nil, "OVERLAY")
	f.DeadText:SetFont(Font, 10, "")
	f.DeadText:SetPoint("CENTER", f, "CENTER", 0, -5)
	f.DeadText:SetWidth(UnitWidth)
	f.DeadText:SetHeight(10)
	f.DeadText:SetTextColor(1, 0, 0, .45)
	f.DeadText:SetShadowColor(.5,.5, .5, .45) 
	f.DeadText:SetShadowOffset(1, -1) 
	f.DeadText:SetJustifyV("TOP")
	f.DeadText:SetJustifyH("CENTER")
	f.DeadText:Hide()

	-- Role icon
	f.RoleIcon = f:CreateTexture()
	f.RoleIcon:ClearAllPoints()
	f.RoleIcon:SetPoint("TOPLEFT", f, "TOPLEFT", 2, -3)
	f.RoleIcon:SetWidth(12)
	f.RoleIcon:SetHeight(12)
	f.RoleIcon:Hide()

	-- Raid icon
	f.RaidIcon = f.Health:CreateTexture()
	f.RaidIcon:SetDrawLayer("OVERLAY", 1)
	f.RaidIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	f.RaidIcon:ClearAllPoints()
	local h = f:GetHeight()/2
	f.RaidIcon:SetWidth(h)
	f.RaidIcon:SetHeight(h)
	f.RaidIcon:SetPoint("CENTER", f.Health, "CENTER")
	f.RaidIcon:Hide()

--	f.ResurrectIcon = f.Health:CreateTexture(nil, "OVERLAY", nil, 4)
--	f.ResurrectIcon::ClearAllPoints()
--	f.ResurrectIcon:SetAllPoints(f.RaidIcon)
	
	-- Debuffs
	f.Debuffs = {}
	local Debuffs = f.Debuffs
	f.DebuffLayer = CreateFrame("Frame", nil, f)
	f.DebuffLayer:SetAllPoints(f)
	f.DebuffLayer:Show()
	
	local Debuff
	for index = 1, DebuffCount do
		Debuffs[index] = CreateFrame("Frame", nil, f.DebuffLayer)
		Debuff = Debuffs[index]
		if not Debuff.SetBackdrop then
			Mixin(Debuff, BackdropTemplateMixin)
		end
		Debuff:EnableMouse(true) 
		Debuff:SetWidth(14)
		Debuff:SetHeight(14)
		Debuff:SetBackdropColor(0, 0, 0, 0)
		Debuff.Image = Debuff:CreateTexture()
		Debuff.Image:SetDrawLayer("ARTWORK", 1)
		Debuff.Image:SetAllPoints()
		if index == 1 then
			Debuff:SetPoint("TOPRIGHT", f.DebuffLayer, "TOPRIGHT", 0 , 3)
		else
			Debuff:SetPoint("RIGHT", Debuffs[index-1], "LEFT")
		end
		Debuff:SetScript("OnEnter", DebuffOnEnter)
		Debuff:SetScript("OnLeave", DebuffOnLeave)
		Debuff.debuffInfo = {}
		Debuff:Hide()	
	end
	
	-- Aggro/Outline Artwork
	f.Aggro = f:CreateTexture() -- behind role icon
	f.Aggro:SetDrawLayer("BORDER", 4)
	f.Aggro:SetTexture(TexturePath.."aggro")
	f.Aggro:SetVertexColor(1,0,0,1)
	f.Aggro:SetBlendMode("ADD")
	f.Aggro:SetAllPoints()
	f.Aggro:Hide()
	
	-- Setup
	SetUnitBackdrop(f)
	SetUnitColors(f)

	-- Scripts
	self:HookScript("OnAttributeChanged", OnChangeAttribute)
	
	self:HookScript("OnEnter", ShowTooltip)
	self:HookScript("OnLeave", HideTooltip)
end

		---------------------- Header Mixin ---------------------
-- Initialization Function: Set Size and other properties of the Unit Button
-- Variables from addons are not accessible when this function is called.
local HeaderSnippet = [[
	local header = self:GetParent()
	self:SetWidth(header:GetAttribute("UnitWidth"))
	self:SetHeight(header:GetAttribute("UnitHeight"))
	header:CallMethod("InitialiseUnit")
]]

local HeaderMixin = {}

function HeaderMixin:InitialiseUnit()
	local UnitFrame = self[#self]
	CreateUnit(UnitFrame)
end

function HeaderMixin:Reset()
	local index = self.index
	if index < 3 then
		self:SetAttribute("UnitWidth", UnitWidth)
		self:SetAttribute("UnitHeight", UnitHeight)
	end
	self:SetAttribute("showRaid", true) -- When true, the group header is shown when the player is in a raid.
	self:SetAttribute("showParty", false) -- When true, the group header is shown when the player is in a party. This attribute doesnt imply showRaid but can work alongside it.
--	self:SetAttribute("showParty", true) -- When true, the group header is shown when the player is in a party. This attribute doesnt imply showRaid but can work alongside it.
	self:SetAttribute("showSolo", Settings.showsolo)
--	self:SetAttribute("showSolo", true)
	
	self:SetAttribute("template", "SecureUnitButtonTemplate")
	self:SetAttribute("templateType", "Button")
	if index == 1 then
		self:SetAttributeByProxy("point", "LEFT")	-- The anchor point per row
	else
		self:SetAttributeByProxy("point", "RIGHT")	-- The anchor point per row
	end
	self:SetAttributeByProxy("columnAnchorPoint", "BOTTOM") -- The anchor point for each new column. A value of LEFT will cause the columns to grow to the right.
	self:SetAttributeByProxy("unitsPerColumn", 5)
	self:SetAttribute("maxColumns", 4)
	if index == 2 then
		self:SetAttribute("xOffset", -xOffset)
	else
		self:SetAttribute("xOffset", xOffset)
	end
	self:SetAttribute("yOffset", yOffset)

--[[
	self:SetAttribute("groupBy", Settings.groupby)
	self:SetAttribute("groupFilter", Settings.groupfilter)
	self:SetAttribute("groupingOrder", Settings.groupingorder)
	self:SetAttribute("maxColumns", Settings.maxcolumns)
	self:SetAttribute("nameList", Settings.namelist)
	self:SetAttribute("roleFilter", Settings.rolefilter)
	self:SetAttribute("sortDir", Settings.sortdir)
	self:SetAttribute("sortMethod", Settings.sortmethod)
	self:SetAttribute("startingIndex", Settings.startingindex)
	self:SetAttribute("strictFiltering", Settings.strictfiltering)

	self:SetAttribute("columnSpacing", 4)
]]--
	self:SetAttribute("initialConfigFunction", HeaderSnippet) -- Init function (Secure)
end

function HeaderMixin:SetAttributeByProxy(name, value) -- clears all raid unit frame points so they can be reset with new settings for these attributes if they change
	if name == "point" or name == "columnAnchorPoint" or name == "unitsPerColumn" then
		local count = 1
		local frame = self:GetAttribute("child" .. count)
		while frame do
			frame:ClearAllPoints()
			count = count + 1
			frame = self:GetAttribute("child" .. count)
		end
	end
	self:SetAttribute(name, value)
end
		---------------------------------------------------------

local function RaidFrameConfigMode(show)
	if Settings.inactive then
		return
	end
	GroupHeaders[1]:SetAttribute("showSolo", show) -- When true, the header is shown when the player is not in any group. This option implies showPlayer.
	if FIZZLEDEBUG then
		NS.DebugPrint("CreateDummyUnits", show, UnitWidth, UnitHeight, xOffset, yOffset, CreateUnit)
	end
end
NS:AddOption("RaidFrameConfigMode", RaidFrameConfigMode)

-- Creates the Group Header, which dynamically controls the creation of Unit Frames
local function CreateGroupHeader(index)
	local points = {
		[1] = {
			[1] = {
				["point"] = "BOTTOMLEFT",
				["anchor"] = 4,
				["topoint"] = "RIGHT",
				["x"] = 6,
				["y"] = 0,
			},
		},
		[2] = {
			[1] = {
				["point"] = "BOTTOMRIGHT",
				["anchor"] = 5,
				["topoint"] = "LEFT",
				["x"] = -6,
				["y"] = 0,
			},
		},
	}

	local GroupHeader = CreateFrame("Frame", "FauzMazzleRF_GroupHeader_" .. index, UIParent, "SecureGroupHeaderTemplate")
	GroupHeader.index = index
	Mixin(GroupHeader, HeaderMixin)
	if Clique then
		GroupHeader:SetAttribute("template", "ClickCastUnitTemplate, SecureUnitButtonTemplate")
		SecureHandlerSetFrameRef(GroupHeader, "clickcast_header", Clique.header)
	else
		GroupHeader:SetAttribute("template", "SecureUnitButtonTemplate")
	end
	local function SetPoints(self, points) -- Also used to anchor Headers
		self:ClearAllPoints()
		local Parent = self:GetParent()
		for i=1, #points do
			local Point = points[i]
			local anchor = Point.anchor == "Parent" and Parent or type(Point.anchor) == "number" and NS:GetFrame("HotSpots", Point.anchor) or Point.anchor and Parent.FrameList[Point.anchor]
			self:SetPoint(Point.point, anchor:GetName() or anchor, Point.topoint, Point.x, Point.y)
		end
	end
	SetPoints(GroupHeader, points[index])
	if not UnitHeight then
		local Cave = NS:GetFrame("Base", "Cave")
		UnitWidth = ceil((Cave:GetLeft() - NS:GetFrame("HotSpots", 4):GetRight())/5)-8
		UnitHeight = floor(Cave:GetHeight() / 4)
		xOffset = 4
		yOffset = 2
		PlayerName = UnitName("player")
		PlayerHighlight:SetSize(UnitHeight/2, UnitHeight/2)
		LeaderHighlight:SetSize(UnitHeight/2, UnitHeight/2)
		LeaderHighlight.LeaderIcon:SetSize(UnitHeight/2.5, UnitHeight/2.5)
	end
	GroupHeader:Reset()

	GroupHeader:Show()
	return GroupHeader
end

------------------------------------
-- Get things started.
------------------------------------
local function InitRaidFrames()
	Settings = NS:GetSettings("ADDONS").RaidFrames1 -- Temp name for settings
	if Settings.inactive then 
		return 
	end
	Settings = NS:GetSettings("RaidFrames1")
	if not Settings.healthcolor then -- initial setup
		return 
	end
	L = NS.TEXTS
	for k, v in pairs(RoleList) do
		v.tip = L[v.tip]
	end
	local Class = select(2, UnitClass("player")) or "ROGUE" 
	PlayerColor = RAID_CLASS_COLORS[Class]

	PlayerHighlight = CreateHighlight(TexturePath.."Corner")

	LeaderHighlight = CreateHighlight(TexturePath.."Corner", "OVERLAY", 1)
	LeaderHighlight:SetFrameLevel(4)
	LeaderHighlight.Texture:SetVertexColor(0.75, 0.55, 0)
	LeaderHighlight.LeaderIcon = LeaderHighlight:CreateTexture()
	LeaderHighlight.LeaderIcon:SetSize(10, 10)
	LeaderHighlight.LeaderIcon:SetBlendMode("ADD")
	LeaderHighlight.LeaderIcon:SetPoint("TOPRIGHT", LeaderHighlight)
	LeaderHighlight.LeaderIcon:SetTexture("Interface\\GroupFrame\\UI-Group-LeaderIcon")

	UpdateSettings()

	------------------------------------
	-- Create the group frames
	------------------------------------
	GroupHeaders[1] = CreateGroupHeader(1)
	GroupHeaders[2] = CreateGroupHeader(2)

	SetOrder()
	GroupHeaders.TrackedAuras = TrackedAuras
--[[
	if CompactRaidFrameContainer then
		CompactRaidFrameContainer:Hide()
		CompactRaidFrameContainer.Show = nil
		CompactRaidFrameContainer:UnregisterAllEvents()
		CompactRaidFrameManager:Hide()
		CompactRaidFrameManager.Hide = nil
		CompactRaidFrameManager:UnregisterAllEvents()
	end
]]--
end

NS:RegisterForInit("PLAYER_ENTERING_WORLD", "InitRaidFrames", 5, InitRaidFrames)
